/*-
 * Copyright (c) 2019 UPLEX Nils Goroll Systemoptimierung
 * All rights reserved
 *
 * Author: Geoffrey Simmons <geoffrey.simmons@uplex.de>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include "config.h"

#include <stdlib.h>
#include <string.h>

#include "cache/cache.h"

/* XXX grr */
#undef ZERO_OBJ
#define ZERO_OBJ(to, sz)      (void)memset(to, 0, sz)

#include "rfc8188.h"
#include "vfp.h"
#include "keys.h"

#define HDR_PREFIX_LEN (SALT_LEN + 4 + 1)
#define ID_LEN 255

#define ERR_DEC(ctx, msg)			\
	VFP_Error((ctx), "ece decrypt: " msg)

#define VERR_DEC(ctx, fmt, ...)					\
	VFP_Error((ctx), "ece decrypt: " fmt, __VA_ARGS__)

struct ece_crypto {
	unsigned		magic;
#define ECE_CRYPTO_MAGIC	0xe7f66e91
	unsigned char		cek[AES128_KEYLEN];
	uint64_t		prenonce_lo;
	uint32_t		prenonce_hi;
};

struct ece {
	unsigned		magic;
#define ECE_MAGIC		0x2b066cec
	struct ece_crypto	*crypto;
	EVP_CIPHER_CTX		*ectx;
	uint8_t			*buf;
	uint64_t		seq_lo;
	uint32_t		seq_hi;
	uint32_t		rs;
};

static enum vfp_status
suck_bytes(struct vfp_ctx *ctx, void *ptr, size_t *lenp)
{
	size_t len, off;
	ssize_t l;
	enum vfp_status vp = VFP_OK;

	CHECK_OBJ_NOTNULL(ctx, VFP_CTX_MAGIC);
	AN(ptr);
	AN(lenp);
	len = *lenp;

	for (off = 0, l = len; vp == VFP_OK && off < len; l = len - l) {
		vp = VFP_Suck(ctx, ptr + off, &l);
		assert(l >= 0);
		assert(vp >= VFP_ERROR && vp < VFP_NULL);
		if (vp == VFP_ERROR)
			return (vp);
		assert((size_t) l <= len - off);
		off += l;
	}
	*lenp = off;
	return (vp);
}

/*
 * RFC8188 ch 2: "The final encoding consists of a header ... and zero or
 * more fixed-size encrypted records ..."
 * So we reject an empty body, but accept a header with no records. The
 * latter results in a zero-length decryption.
 */
static enum vfp_status
decrypt_init(struct ece *ece, struct vfp_ctx *ctx)
{
	size_t len;
	uint8_t idlen, *key, hdr[HDR_PREFIX_LEN], id[ID_LEN], prk[SHA256_LEN],
		cek[SHA256_LEN], prenonce[NONCE_LEN];
	char errmsg[ERRMSG_LEN];
	enum vfp_status vp;

	CHECK_OBJ_NOTNULL(ece, ECE_MAGIC);
	CHECK_OBJ_NOTNULL(ece->crypto, ECE_CRYPTO_MAGIC);
	CHECK_OBJ_NOTNULL(ctx, VFP_CTX_MAGIC);
	AZ(ece->rs);
	AZ(ece->seq_hi);
	AZ(ece->seq_lo);

	len = HDR_PREFIX_LEN;
	vp = suck_bytes(ctx, hdr, &len);
	if (vp == VFP_ERROR)
		return (vp);
	if (vp == VFP_END) {
		if (len == 0)
			return (ERR_DEC(ctx, "empty body"));
		else
			return (ERR_DEC(ctx, "truncated header"));
	}

	decode_header(hdr, &ece->rs, &idlen);
	if (ece->rs < 18)
		return(VERR_DEC(ctx, "invalid record size %u", ece->rs));

	errno = 0;
	ece->buf = malloc(ece->rs);
	if (ece->buf == NULL)
		return (VERR_DEC(ctx, "buffer allocation failed: %s",
				 strerror(errno)));

	len = idlen;
	vp  = suck_bytes(ctx, id, &len);
	if (vp == VFP_ERROR)
		return (vp);
	if (vp == VFP_END) {
		if (len == idlen)
			/* Empty decryption */
			return (vp);
		else
			return (ERR_DEC(ctx, "truncated header"));
	}

	/* XXX rdlock key table */
	key = get_key(id, idlen);
	if (key == NULL)
		vp = VERR_DEC(ctx, "unknown key %.*s", idlen, id);
	else if (derive_prk(hdr, key, prk, errmsg) != 0)
		vp = VERR_DEC(ctx, "%s", errmsg);
	/* XXX rdunlock key table */

	if (vp == VFP_ERROR)
		return (vp);
	if (derive_cek(prk, cek, errmsg) != 0
	    || derive_prenonce(prk, prenonce, errmsg) != 0)
		return (VERR_DEC(ctx, "%s", errmsg));

	ece->crypto->prenonce_hi = vbe32dec(prenonce);
	ece->crypto->prenonce_lo = vbe64dec(prenonce + 4);

	memcpy(ece->crypto->cek, cek, AES128_KEYLEN);
	return (VFP_OK);
}

static inline void
seq_inc(struct ece *ece)
{
	CHECK_OBJ_NOTNULL(ece, ECE_MAGIC);

	if (ece->seq_lo < UINT64_MAX)
		ece->seq_lo++;
	else {
		assert(ece->seq_hi != UINT32_MAX);
		ece->seq_hi++;
		ece->seq_lo = 0;
	}
}

static inline void
nonce_xor_seq(struct ece *ece, uint8_t *nonce)
{
	uint32_t nonce_hi;
	uint64_t nonce_lo;

	CHECK_OBJ_NOTNULL(ece, ECE_MAGIC);
	CHECK_OBJ_NOTNULL(ece->crypto, ECE_CRYPTO_MAGIC);
	AN(nonce);

	nonce_hi = ece->crypto->prenonce_hi ^ ece->seq_hi;
	nonce_lo = ece->crypto->prenonce_lo ^ ece->seq_lo;

	vbe32enc(nonce, nonce_hi);
	vbe64enc(nonce + 4, nonce_lo);
}

static enum vfp_status
decrypt(struct ece *ece, struct vfp_ctx *ctx, unsigned char *plaintext,
	size_t *lenp, enum vfp_status vp)
{
	ssize_t ciphertext_len, plaintext_len;
	unsigned char nonce[NONCE_LEN], *tag;
	char errmsg[ERRMSG_LEN];
	int last;

	CHECK_OBJ_NOTNULL(ece, ECE_MAGIC);
	CHECK_OBJ_NOTNULL(ece->crypto, ECE_CRYPTO_MAGIC);

	if (*lenp < 18)
		return (ERR_DEC(ctx, "record truncated"));
	ciphertext_len = *lenp - TAG_LEN;
	tag = ece->buf + ciphertext_len;

	nonce_xor_seq(ece, nonce);
	plaintext_len = decrypt_record(ece->ectx, ece->buf, ciphertext_len,
				       tag, ece->crypto->cek, nonce, plaintext,
				       &last, errmsg);

	if (plaintext_len < 0)
		return (VERR_DEC(ctx, "%s", errmsg));
	if (last && vp != VFP_END)
		return (ERR_DEC(ctx, "premature last record"));
	if (!last && vp == VFP_END)
		return (ERR_DEC(ctx, "message truncated"));
	if (!last && *lenp < ece->rs)
		return (ERR_DEC(ctx, "record smaller than record size"));

	seq_inc(ece);
	*lenp = plaintext_len;
	return (vp);
}

/* VFP interface */

static enum vfp_status v_matchproto_(vfp_init_f)
vfp_decrypt_init(struct vfp_ctx *ctx, struct vfp_entry *ent)
{
	struct ece *ece;
	struct ece_crypto *crypto;
	char errmsg[ERRMSG_LEN];

	CHECK_OBJ_NOTNULL(ctx, VFP_CTX_MAGIC);
	CHECK_OBJ_NOTNULL(ent, VFP_ENTRY_MAGIC);
	AN(ent->vfp);

	/* XXX implement me */
	if (http_GetStatus(ctx->resp) == 206)
	 	return (VFP_NULL);

	if (!http_HdrIs(ctx->resp, H_Content_Encoding, "aes128gcm"))
		return (VFP_NULL);

	errno = 0;
	ALLOC_OBJ(ece, ECE_MAGIC);
	if (ece == NULL)
		return (VERR_DEC(ctx, "allocation in init: %s",
				 strerror(errno)));

	ece->ectx = cipher_ctx_init(0, errmsg);
	if (ece->ectx == NULL) {
		FREE_OBJ(ece);
		return (VERR_DEC(ctx, "initializing cipher context: %s",
				 errmsg));
	}

	ALLOC_OBJ(crypto, ECE_CRYPTO_MAGIC);
	if (crypto == NULL) {
		FREE_OBJ(ece);
		return (VERR_DEC(ctx, "allocation in init: %s",
				 strerror(errno)));
	}
	ece->crypto = crypto;
	ent->priv1 = ece;

	http_Unset(ctx->resp, H_Content_Encoding);
	http_Unset(ctx->resp, H_Content_Length);
	RFC2616_Weaken_Etag(ctx->resp);
	ctx->obj_flags |= OF_CHGCE;

	return (VFP_OK);
}

static void v_matchproto_(vfp_fini_f)
vfp_decrypt_fini(struct vfp_ctx *ctx, struct vfp_entry *ent)
{
	struct ece *ece;

	CHECK_OBJ_NOTNULL(ctx, VFP_CTX_MAGIC);
	CHECK_OBJ_NOTNULL(ent, VFP_ENTRY_MAGIC);

	if (ent->priv1 == NULL)
		return;

	CAST_OBJ(ece, ent->priv1, ECE_MAGIC);
	if (ece->ectx != NULL)
		cipher_ctx_fini(ece->ectx);
	if (ece->buf != NULL)
		free(ece->buf);
	if (ece->crypto == NULL)
		return;
	CHECK_OBJ(ece->crypto, ECE_CRYPTO_MAGIC);
	FREE_OBJ(ece->crypto);
}

static enum vfp_status v_matchproto_(vfp_pull_f)
vfp_decrypt_pull(struct vfp_ctx *ctx, struct vfp_entry *ent, void *ptr,
		 ssize_t *lenp)
{
	struct ece *ece;
	size_t len;
	enum vfp_status vp;

	CHECK_OBJ_NOTNULL(ctx, VFP_CTX_MAGIC);
	CHECK_OBJ_NOTNULL(ent, VFP_ENTRY_MAGIC);
	CAST_OBJ_NOTNULL(ece, ent->priv1, ECE_MAGIC);
	AN(ptr);
	AN(lenp);

	*lenp = 0;
	if (ece->rs == 0)
		if ((vp = decrypt_init(ece, ctx)) != VFP_OK)
			return (vp);
	AN(ece->rs);
	AN(ece->buf);
	len = ece->rs;
	vp = suck_bytes(ctx, ece->buf, &len);
	assert(len <= ece->rs);
	if (vp == VFP_ERROR)
		return (vp);
	vp = decrypt(ece, ctx, ptr, &len, vp);
	*lenp = len;
	return (vp);
}

const struct vfp vfp_decrypt = {
	.name = "ece_decrypt",
	.init = vfp_decrypt_init,
	.pull = vfp_decrypt_pull,
	.fini = vfp_decrypt_fini,
};
